#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <pci/pci.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <setjmp.h>

#include "haps.h"
#include "mem.h"
#include "md5.h"
#include "list.h"
#include "module.h"

/* Cardbus stuff can show up as PCI. */
/* Start at 1 so we don't do weird stuff on previous installations. */
#define PCI_UNKNOWN		0
#define PCI_NORMAL		1
#define PCI_CARDBUS		2

enum deviceClass
{
  /* device classes... this is somewhat ad-hoc */
  CLASS_UNSPEC, CLASS_OTHER, CLASS_NETWORK, CLASS_SCSI, CLASS_MOUSE,
  CLASS_AUDIO, CLASS_CDROM, CLASS_MODEM, CLASS_VIDEO, CLASS_TAPE,
  CLASS_FLOPPY, CLASS_SCANNER, CLASS_HD, CLASS_RAID, CLASS_PRINTER,
  CLASS_CAPTURE, CLASS_KEYBOARD, CLASS_MONITOR, CLASS_USB,
  CLASS_SOCKET, CLASS_FIREWIRE
};

struct pciDevice
{
  /* common fields */
  struct device *next;		/* next device in list */
  int index;
  enum deviceClass class;	/* type */
//  enum deviceBus bus;         /* bus it's attached to */
  char *device;			/* device file associated with it */
  char *driver;			/* driver to load, if any */
  char *desc;			/* a description */
  int detached;
  /* pci-specific fields */
  unsigned int vendorId;	/* vendor id */
  unsigned int deviceId;	/* device id */
  int pciType;			/* type of PCI bus */
  unsigned int subVendorId;
  unsigned int subDeviceId;
};

struct SerialPort
{
  int index;
  unsigned long port;
  unsigned long irq;
};

static struct SerialPort *serials;
static int num_serials = -1;

static struct pciDevice *pciDeviceList = NULL;
static int numPciDevices = 0;
static struct pci_access *pacc = NULL;
static jmp_buf pcibuf;

char *
pciToKudzu (unsigned int class)
{

  if (!class)
    return CLASS_UNSPEC;
  switch (class >> 8)
  {
  case PCI_BASE_CLASS_DISPLAY:
    return "VIDEO";
  default:
    break;
  }
  switch (class)
  {
  case PCI_CLASS_STORAGE_SCSI:
  case PCI_CLASS_SERIAL_FIBER:
    return "SCSI";
  case PCI_CLASS_STORAGE_FLOPPY:
    return "FLOPPY";
  case PCI_CLASS_STORAGE_RAID:
    return "RAID";
  case PCI_CLASS_MULTIMEDIA_AUDIO:
    return "AUDIO";
  case PCI_CLASS_INPUT_MOUSE:
    return "MOUSE";
  case PCI_CLASS_MULTIMEDIA_VIDEO:
    return "CAPTURE";
  case PCI_CLASS_COMMUNICATION_SERIAL:
  case PCI_CLASS_COMMUNICATION_OTHER:
    return "MODEM";
  case PCI_CLASS_NOT_DEFINED_VGA:
    return "VIDEO";
    /* Fix for one of the megaraid variants.
     * It claims to be an I2O controller. */
  case 0x0e00:
    return "SCSI";
  case PCI_CLASS_SERIAL_USB:
    return "USB";
  case PCI_CLASS_SERIAL_FIREWIRE:
    return "FIREWIRE";
  case PCI_CLASS_BRIDGE_CARDBUS:
    return "SOCKET";

  case PCI_CLASS_NETWORK_ETHERNET:
  case PCI_CLASS_NETWORK_TOKEN_RING:
  case PCI_CLASS_NETWORK_FDDI:
  case PCI_CLASS_NETWORK_ATM:
    return "NETWORK";

  default:
    return "OTHER";
  }
}

static void
InitSerials ()
{
  FILE *fp;
  char buffer[256];
  char *uart;
  char *port;
  char *irq;
  char *sp;
  int i;

  fp = fopen ("/proc/tty/driver/serial", "r");
  if (!fp)
    return;

  num_serials = 0;
  while (fgets (buffer, sizeof (buffer), fp))
  {
    if (strcasestr (buffer, "uart"))
      num_serials++;
  }

  if (!num_serials)
  {
    fclose (fp);
    return;
  }

  serials = malloc (sizeof (struct SerialPort) * num_serials);
  if (!serials)
  {
    perror ("malloc");
    num_serials = 0;
    fclose (fp);
    return;
  }

  i = 0;
  rewind (fp);
  while (i < num_serials && fgets (buffer, sizeof (buffer), fp))
  {
    uart = strcasestr (buffer, "uart");
    if (uart)
    {
      port = strcasestr (buffer, "port");
      irq = strcasestr (buffer, "irq");
      uart = strchr (uart, ':');
      uart++;
      sp = strchr (uart, ' ');
      *sp = '\0';
      serials[i].index = serials[i].port = serials[i].irq = 0;
      if (strcasecmp (uart, "unknown") && port && irq)
      {
	serials[i].index = strtol (buffer, NULL, 10);
	serials[i].port = strtoul (port + 5, NULL, 16);
	serials[i].irq = strtoul (irq + 4, NULL, 10);
	i++;
      }
    }
  }
  num_serials = i;
  fclose (fp);
}

void
checkPCISerial (struct node *pci, struct pci_dev *dev)
{
  unsigned long io_base[6];
  int i, j;
  char buffer[256];

  if (num_serials == -1)
    InitSerials ();

  if (num_serials == 0)
    return;

  /* Find out our io ports. */
  for (i = 0; i < 6; i++)
  {
    io_base[i] = dev->base_addr[i];
    if (io_base[i] & PCI_BASE_ADDRESS_SPACE_IO)
      io_base[i] &= PCI_BASE_ADDRESS_IO_MASK;
  }

  /* Match irq and io ports. */
  for (i = 0; i < num_serials; i++)
  {
    /* Check irq first. */
    if (serials[i].irq != dev->irq)
      continue;

    /* Check io ports. */
    for (j = 0; j < 6; j++)
    {
      if (io_base[j] <= serials[i].port
	  && serials[i].port <= io_base[j] + 127)
      {
	snprintf (buffer, sizeof (buffer), "ttyS%d", serials[i].index);
	nodesetprop (pci, "DEVICE", buffer);
	return;
      }
    }
  }
}

static int
vendCmp (const void *a, const void *b)
{
  const struct pciDevice *one = a;
  const struct pciDevice *two = b;

  return (one->vendorId - two->vendorId);
}

static char *
getVendor (unsigned int vendor)
{
  struct pciDevice *searchDev, key;
  char *tmpstr;

  key.vendorId = vendor;

  searchDev = bsearch (&key, pciDeviceList, numPciDevices,
		       sizeof (struct pciDevice), vendCmp);
  if (searchDev)
  {
    int x;

    x = strchr (searchDev->desc, '|') - searchDev->desc;
    if (x <= 0 || !(tmpstr = calloc (x + 1, sizeof (char))))
      return NULL;
    tmpstr = strncpy (tmpstr, searchDev->desc, x);
    tmpstr[x] = '\0';
    return tmpstr;
  }
  else
  {
    return NULL;
  }
}

static int
devCmp (const void *a, const void *b)
{
  const struct pciDevice *one = a;
  const struct pciDevice *two = b;
  int x = 0, y = 0, z = 0, xx = 0, yy = 0;

  x = (one->vendorId - two->vendorId);
  xx = (one->subVendorId - two->subVendorId);
  y = (one->deviceId - two->deviceId);
  yy = (one->subDeviceId - two->subDeviceId);
  if (one->pciType && two->pciType)
    z = (one->pciType - two->pciType);
  if (x)
    return x;
  else if (y)
    return y;
  else if (xx)
    return xx;
  else if (yy)
    return yy;
  else
    return z;
}

int
pciReadDrivers (char *filename)
{
  int fd;
  char *buf;
  int numDrivers;
  int vendid, devid, subvendid, subdevid;
  int merge = 0;
  char *start, *ptr, *bufptr;
  struct pciDevice *nextDevice, *tmpdev = NULL, key;
  char *module;
  char *descrip;
  char path[256];
  struct utsname utsbuf;


  if (filename)
  {
    fd = open (filename, O_RDONLY);
    if (fd < 0)
      return -1;
  }
  else
  {
    fd = open (PCITABLE, O_RDONLY);
    if (fd < 0)
    {
      return -1;
    }
  }
  buf = bufFromFd (fd);

  /* upper bound */
  numDrivers = 1;
  start = buf;
  while ((start = strchr (start, '\n')))
  {
    numDrivers++;
    start++;
  }

  if (pciDeviceList)
    merge = 1;

  pciDeviceList = realloc (pciDeviceList, sizeof (*pciDeviceList) *
			   (numPciDevices + numDrivers));
  nextDevice = pciDeviceList + numPciDevices;

  start = buf;
  while (start && *start)
  {
    while (isspace (*start))
      start++;
    if (*start != '#' && *start != '\n')
    {
      vendid = strtoul (start, &ptr, 16);
      if (!(*ptr && *ptr != '\n'))
	continue;
      start = ptr + 1;
      devid = strtoul (start, &ptr, 16);
      if (!(*ptr && *ptr != '\n'))
	continue;
      start = ptr + 1;
      subvendid = strtoul (start, &ptr, 16);
      if (start != ptr)
      {
	if (!(*ptr && *ptr != '\n'))
	  continue;
	start = ptr + 1;
	subdevid = strtoul (start, &ptr, 16);
      }
      else
      {
	subvendid = 0xffff;
	subdevid = 0;
      }
      while (*ptr && *ptr != '"' && *ptr != '\n')
	ptr++;
      if (*ptr != '"')
	continue;
      ptr++;
      start = ptr;
      while (*ptr && *ptr != '"' && *ptr != '\n')
	ptr++;
      if (*ptr != '"')
	continue;
      *ptr = '\0';
      module = strdup (start);
      ptr++;
      while (*ptr && *ptr != '"' && *ptr != '\n')
	ptr++;
      if (*ptr != '"')
	continue;
      ptr++;
      start = ptr;
      while (*ptr && *ptr != '"' && *ptr != '\n')
	ptr++;
      if (*ptr != '"')
	continue;
      *ptr = '\0';
      descrip = strdup (start);
      ptr++;
      start = ptr;

      if (merge)
      {
	tmpdev = nextDevice;
	key.vendorId = vendid;
	key.deviceId = devid;

	key.subVendorId = subvendid;
	key.subDeviceId = subdevid;
	if (strncmp (module, "CardBus:", 8) == 0)
	  key.pciType = PCI_CARDBUS;
	else
	  key.pciType = PCI_NORMAL;

	nextDevice = bsearch (&key, pciDeviceList, numPciDevices,
			      sizeof (struct pciDevice), devCmp);
	if (!nextDevice)
	{
	  nextDevice = tmpdev;
	  tmpdev = NULL;
	  numPciDevices++;
	}
	else
	{
	  if (nextDevice->device)
	    free (nextDevice->device);
	  if (nextDevice->desc)
	    free (nextDevice->desc);
	  if (nextDevice->driver)
	    free (nextDevice->driver);
	}
      }
      else
	numPciDevices++;
      nextDevice->vendorId = vendid;
      nextDevice->deviceId = devid;
      nextDevice->subVendorId = subvendid;
      nextDevice->subDeviceId = subdevid;
      if (strncmp (module, "CardBus:", 8) == 0)
      {
	nextDevice->pciType = PCI_CARDBUS;
	nextDevice->driver = strdup (&module[8]);
      }
      else
      {
	nextDevice->pciType = PCI_NORMAL;
	nextDevice->driver = strdup (module);
      }
      nextDevice->desc = strdup (descrip);
      nextDevice->next = NULL;
      nextDevice->device = NULL;
      nextDevice->class = 0;
      if (merge && tmpdev)
	nextDevice = tmpdev;
      else
      {
	nextDevice++;
	if (merge)
	  qsort (pciDeviceList, numPciDevices, sizeof (*pciDeviceList),
		 devCmp);
      }
      free (module);
      free (descrip);
    }

    while (*start && *start != '\n')
      start++;
    if (start)
      start++;
  }
  free (buf);

  qsort (pciDeviceList, numPciDevices, sizeof (*pciDeviceList), devCmp);

  /* OK, now read the modutils modules.pcimap file. */
  if (!filename)
  {
    uname (&utsbuf);
    if (strstr (utsbuf.release, "BOOT"))
    {
      char kernelver[64];
      int len;

      len = strstr (utsbuf.release, "BOOT") - utsbuf.release;
      strncpy (kernelver, utsbuf.release, len);
      kernelver[len] = '\0';
      snprintf (path, 255, "/lib/modules/%s/modules.pcimap", kernelver);
    }
    else
    {
      snprintf (path, 255, "/lib/modules/%s/modules.pcimap", utsbuf.release);
    }
    fd = open (path, O_RDONLY);
    if (fd < 0)
    {
      fd = open ("/etc/modules.pcimap", O_RDONLY);
      if (fd < 0)
      {
	fd = open ("/modules/modules.pcimap", O_RDONLY);
	if (fd < 0)
	{
	  fd = open ("./modules.pcimap", O_RDONLY);
	  if (fd < 0)
	    return 0;
	}
      }
    }
    bufptr = bufFromFd (fd);

    start = buf = bufptr;
    numDrivers = 0;
    while ((start = strchr (start, '\n')))
    {
      numDrivers++;
      start++;
    }
    pciDeviceList = realloc (pciDeviceList, sizeof (*pciDeviceList) *
			     (numPciDevices + numDrivers));
    nextDevice = pciDeviceList + numPciDevices;

    start = buf = bufptr;
    while (start && *start)
    {
      while (*buf && *buf != '\n')
	buf++;
      if (*buf)
      {
	*buf = '\0';
	buf++;
      }
      if (*start == '#')
      {
	start = buf;
	continue;
      }
      ptr = start;
      while (*ptr && *ptr != ' ')
	ptr++;
      if (*ptr)
      {
	*ptr = '\0';
	ptr++;
      }
      module = strdup (start);
      while (*ptr && *ptr == ' ')
	ptr++;
      if (!*ptr)
      {
	start = buf;
	free (module);
	continue;
      }
      start = ptr;
      while (*ptr && *ptr != ' ')
	ptr++;
      if (*ptr)
      {
	*ptr = '\0';
	ptr++;
      }
      vendid = strtoul (start, (char **) NULL, 16);
      if (vendid == 0xffffffff)
      {
	start = buf;
	free (module);
	continue;
      }
      while (*ptr && *ptr == ' ')
	ptr++;
      if (!*ptr)
      {
	start = buf;
	free (module);
	continue;
      }
      start = ptr;
      while (*ptr && *ptr != ' ')
	ptr++;
      if (*ptr)
      {
	*ptr = '\0';
	ptr++;
      }
      devid = strtoul (start, (char **) NULL, 16);
      while (*ptr && *ptr == ' ')
	ptr++;
      if (!*ptr)
      {
	start = buf;
	free (module);
	continue;
      }
      start = ptr;
      while (*ptr && *ptr != ' ')
	ptr++;
      if (*ptr)
      {
	*ptr = '\0';
	ptr++;
      }
      subvendid = strtoul (start, (char **) NULL, 16);
      if (subvendid == 0xffffffff)
      {
	subvendid = 0xffff;
	subdevid = 0;
      }
      else
      {
	while (*ptr && *ptr == ' ')
	  ptr++;
	if (!*ptr)
	{
	  start = buf;
	  free (module);
	  continue;
	}
	start = ptr;
	while (*ptr && *ptr != ' ')
	  ptr++;
	if (*ptr)
	{
	  *ptr = '\0';
	  ptr++;
	}
	subdevid = strtoul (start, (char **) NULL, 16);
	if (subdevid == 0xffffffff)
	  subdevid = 0;
      }
      start = buf;

      tmpdev = nextDevice;
      key.vendorId = vendid;
      key.deviceId = devid;
      key.subVendorId = subvendid;
      key.subDeviceId = subdevid;
      key.pciType = PCI_NORMAL;
      nextDevice = bsearch (&key, pciDeviceList, numPciDevices,
			    sizeof (struct pciDevice), devCmp);
      /* HACK */
      if (nextDevice && !strcmp (nextDevice->driver, "bcm5700"))
	nextDevice = NULL;
      /* HACK */
      if (nextDevice && !strcmp (nextDevice->driver, "eepro100"))
	nextDevice = NULL;
      if (!nextDevice)
      {
	key.vendorId = vendid;
	key.deviceId = devid;
	key.subVendorId = 0xffff;
	key.subDeviceId = 0;
	key.pciType = PCI_NORMAL;
	nextDevice = bsearch (&key, pciDeviceList, numPciDevices,
			      sizeof (struct pciDevice), devCmp);
      }
      if (!nextDevice)
      {
	char newdesc[64];
	nextDevice = tmpdev;
	tmpdev = NULL;
	numPciDevices++;
	nextDevice->vendorId = vendid;
	nextDevice->deviceId = devid;
	nextDevice->subVendorId = subvendid;
	nextDevice->subDeviceId = subdevid;
	nextDevice->pciType = PCI_NORMAL;
	nextDevice->driver = strdup (module);
	snprintf (newdesc, 63, "Unknown vendor|Generic %s device", module);
	nextDevice->desc = strdup (newdesc);
	nextDevice->next = NULL;
	nextDevice->device = NULL;
	nextDevice->class = 0;
      }
      else
      {
	if (!strcmp (nextDevice->driver, "unknown") ||
	    !strcmp (nextDevice->driver, "disabled") ||
	    !strcmp (nextDevice->driver, "ignore"))
	{
	  free (nextDevice->driver);
	  nextDevice->driver = strdup (module);
	}
      }
      if (tmpdev)
	nextDevice = tmpdev;
      else
      {
	nextDevice++;
	qsort (pciDeviceList, numPciDevices, sizeof (*pciDeviceList), devCmp);
      }
      free (module);
    }
    free (bufptr);
  }
  return 0;
}

void
pciFreeDrivers (void)
{
  int x;

  if (pciDeviceList)
  {
    for (x = 0; x < numPciDevices; x++)
    {
      if (pciDeviceList[x].device)
	free (pciDeviceList[x].device);
      if (pciDeviceList[x].driver)
	free (pciDeviceList[x].driver);
      if (pciDeviceList[x].desc)
	free (pciDeviceList[x].desc);
    }
    free (pciDeviceList);
    pciDeviceList = NULL;
    numPciDevices = 0;
  }
}

struct node *
pciGetDeviceInfo (unsigned int vend, unsigned int dev,
		  unsigned int subvend, unsigned int subdev, int bus)
{
  struct pciDevice *searchDev, key;
  struct node *device;
  char buf[0x100];

  device = nodealloc ();
  nodesetprop (device, "BUS", "PCI");
  nodesetpropint (device, "PCIBUS", bus);

  key.vendorId = vend;
  key.deviceId = dev;
  key.pciType = bus;
  key.subVendorId = subvend;
  key.subDeviceId = subdev;

  searchDev = bsearch (&key, pciDeviceList, numPciDevices,
		       sizeof (struct pciDevice), devCmp);
  if (!searchDev && key.pciType != PCI_NORMAL)
  {
    key.pciType = PCI_NORMAL;
    searchDev = bsearch (&key, pciDeviceList, numPciDevices,
			 sizeof (struct pciDevice), devCmp);
  }
  if (!searchDev)
  {
    key.pciType = bus;
    key.subVendorId = 0xffff;
    key.subDeviceId = 0;
    searchDev = bsearch (&key, pciDeviceList, numPciDevices,
			 sizeof (struct pciDevice), devCmp);
  }
  if (!searchDev /*&& key.pciType != PCI_NORMAL */ )
  {
    key.pciType = PCI_NORMAL;
    searchDev = bsearch (&key, pciDeviceList, numPciDevices,
			 sizeof (struct pciDevice), devCmp);
  }
  if (!searchDev)
  {
    char *namebuf;
    nodesetpropint (device, "VENDORID", vend);
    nodesetpropint (device, "DEVICEID", dev);
    nodesetpropint (device, "SUBVENDORID", subvend);
    nodesetpropint (device, "SUBDEVICEID", subdev);
    nodesetprop (device, "DRIVER", "unknown");

/*
    namebuf = getVendor (vend);
    if (!namebuf)
    {
      snprintf (buf, 128,
		"Unknown vendor|unknown device %04x:%04x",
		searchDev->vendorId, searchDev->deviceId);
      nodesetprop (device, "DESC", buf);
    }
    else
    {
      snprintf (buf, 128,
		"%s|unknown device %04x:%04x",
		namebuf, searchDev->vendorId, searchDev->deviceId);
      nodesetprop (device, "DESC", buf);
      free (namebuf);
    }
*/

  }
  else
  {
/*    searchDev = pciNewDevice (searchDev);
    searchDev->pciType = bus;
    searchDev->subVendorId = subvend;
    searchDev->subDeviceId = subdev;
*/
    nodesetpropint (device, "VENDORID", searchDev->vendorId);
    nodesetpropint (device, "DEVICEID", searchDev->deviceId);
    nodesetprop (device, "DRIVER", searchDev->driver);

    nodesetpropint (device, "SUBVENDORID", subvend);
    nodesetpropint (device, "SUBDEVICEID", subdev);
  }
  return device;
}

static void
pcinull (char *foo, ...)
{
}

static void
pcibail (char *foo, ...)
{
  longjmp (pcibuf, 1);
}

static int
isDisabled (struct pci_dev *p, byte config[256])
{
  int disabled;
  int i;
#ifdef __i386__
  int limit = 6;
#else
  int limit = 1;
#endif
  unsigned int devtype, command;

  devtype = config[PCI_CLASS_DEVICE + 1] << 8 | config[PCI_CLASS_DEVICE];
  /* Network, USB, and audio devices with no IRQ are
   * generally 'disabled-in-bios' */
  /* if ( (p->irq == 0 || p->irq == 255) &&
   *   (pciToKudzu(devtype) == CLASS_NETWORK ||
   *    pciToKudzu(devtype) == CLASS_USB ||
   *    pciToKudzu(devtype) == CLASS_FIREWIRE ||
   *    pciToKudzu(devtype) == CLASS_AUDIO
   *    )) {
   *      return 1;
   *}
   */
  if (p->irq)
  {
    return 0;
  }
  /* Ditto for things that are explicitly disabled */
  command = config[PCI_COMMAND];
  disabled = 0;
  for (i = 0; i < limit; i++)
  {
    int x = PCI_BASE_ADDRESS_0 + 4 * i;
    pciaddr_t pos = p->base_addr[i];
    pciaddr_t len = (p->known_fields & PCI_FILL_SIZES) ? p->size[i] : 0;
    u32 flag =
      config[x + 3] << 24 | config[x + 2] << 16 | config[x +
							 1] << 8 | config[x];
    if ((flag == 0xffffffff || !flag) && !pos && !len)
      continue;
    disabled++;
    if ((flag & PCI_BASE_ADDRESS_SPACE_IO) && (command & PCI_COMMAND_IO))
      disabled--;
    else if (command & PCI_COMMAND_MEMORY)
      disabled--;
  }
  return disabled;
}

struct node *
probe (struct node *devices, int params)
{
  struct pci_dev *p;
  /* This should be plenty. */
  int cardbus_bridges[32];
  int bridgenum = 0;
  int init_list = 0;
//  unsigned int type = kudzuToPci (probeClass), devtype;
  unsigned int devtype;
  MD5_CTX context;
  unsigned char digest[16];
  char buf[0x100],buf2[0x100];
  int i;


//  struct node *newdev;

  devices = nodecondremove (devices, "BUS", "PCI", 1);

//  newdev = nodealloc ();
//  nodesetprop (newdev, "BUS", "PCI");
//  nodesetprop (newdev, "CONN", "pci.");
//  nodesetprop (newdev, "CLASS", "BIOS");

  pacc = pci_alloc ();
  if (!pacc)
    return devices;

  if (!pciDeviceList)
  {
    pciReadDrivers (NULL);
    init_list = 1;
  }
  pacc->debug = pacc->warning = pcinull;
  pacc->error = pcibail;
  if (!setjmp (pcibuf))
  {
//    int order = 0;

    pci_init (pacc);
    pci_scan_bus (pacc);

    memset (cardbus_bridges, '\0', 32);
    /* enumerate cardbus bridges first */
    for (p = pacc->devices; p; p = p->next)
    {
      byte config[256];

      memset (config, '\0', 256);
      pci_read_block (p, 0, config, 64);
      if ((config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
      {
	/* Cardbus bridge */
	pci_read_block (p, 64, config + 64, 64);
	for (bridgenum = 0; cardbus_bridges[bridgenum]; bridgenum++);
	cardbus_bridges[bridgenum] = config[PCI_CB_CARD_BUS];
      }
    }

    for (p = pacc->devices; p; p = p->next)
    {
      byte config[256];
      int bustype;
      unsigned int subvend, subdev;
      struct node *dev;

      memset (config, '\0', 256);
      pci_read_block (p, 0, config, 64);
      if ((config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
      {
	/* Cardbus bridge */
	pci_read_block (p, 64, config + 64, 64);
	subvend =
	  config[PCI_CB_SUBSYSTEM_VENDOR_ID +
		 1] << 8 | config[PCI_CB_SUBSYSTEM_VENDOR_ID];
	subdev =
	  config[PCI_CB_SUBSYSTEM_ID + 1] << 8 | config[PCI_CB_SUBSYSTEM_ID];
      }
      else
      {
	subvend =
	  config[PCI_SUBSYSTEM_VENDOR_ID +
		 1] << 8 | config[PCI_SUBSYSTEM_VENDOR_ID];
	subdev = config[PCI_SUBSYSTEM_ID + 1] << 8 | config[PCI_SUBSYSTEM_ID];
      }
      bustype = PCI_NORMAL;
      for (bridgenum = 0; cardbus_bridges[bridgenum]; bridgenum++)
      {
	if (p->bus == cardbus_bridges[bridgenum])
	  bustype = PCI_CARDBUS;
      }
      dev =
	pciGetDeviceInfo (p->vendor_id, p->device_id, subvend, subdev,
			  bustype);


/*
    namebuf = getVendor (vend);
    if (!namebuf)
    {
      snprintf (buf, 128,
		"Unknown vendor|unknown device %04x:%04x",
		searchDev->vendorId, searchDev->deviceId);
      nodesetprop (device, "DESC", buf);
    }
    else
    {
      snprintf (buf, 128,
		"%s|unknown device %04x:%04x",
		namebuf, searchDev->vendorId, searchDev->deviceId);
      nodesetprop (device, "DESC", buf);
      free (namebuf);
    }
*/

      sprintf(buf,"pci.%d.%d.%d",p->bus,p->dev,p->func);
      nodesetprop(dev,"CONN",buf);

      nodesetprop(dev,"PCICLASS",pci_lookup_name(p->access, buf, sizeof(buf), PCI_LOOKUP_CLASS,
			 config[PCI_CLASS_DEVICE]+(config[PCI_CLASS_DEVICE+1]<<8), 0, 0, 0));
      
      nodesetprop(dev,"DESC",pci_lookup_name(p->access, buf, sizeof(buf),
			 PCI_LOOKUP_DEVICE,
			 p->vendor_id, p->device_id, 0, 0));
      
      nodesetprop(dev,"VENDOR",pci_lookup_name(p->access, buf, sizeof(buf), PCI_LOOKUP_VENDOR,
			 p->vendor_id, p->device_id, 0, 0));

      nodesetpropint (dev, "PCIDEV", p->dev);
      nodesetpropint (dev, "PCIFUNC", p->func);
      nodesetpropint (dev, "KNOWN_FIELDS", p->known_fields);
      nodesetpropint (dev, "IRQ", p->irq);

      nodesetprophint (dev, "ROM_BASE_ADDR", p->rom_base_addr);
      nodesetprophint (dev, "ROM_SIZE", p->rom_size);
      memmd5(buf, p->rom_base_addr,p->rom_size);
      nodesetprop (dev, "MD5_ROM", buf);

/*
      MD5Init (&context);
      for (i=0;i<6;i++)      
      {
        sprintf(buf2,"BASE_ADDR_MEM_%d",i);
        nodesetprophint (dev, buf2, p->base_addr[i]);
        sprintf(buf2,"SIZE_MEM_%d",i);
        nodesetprophint (dev, buf2, p->size[i]);
        sprintf(buf2,"MD5_MEM_%d",i);
        memmd5(buf, p->base_addr[i],p->size[i]);
        nodesetprop (dev, buf2, buf);
        MD5UpdateS (&context,buf);
      }
      MD5Final (digest, &context);

      sprintf (buf,
	"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
	digest[0], digest[1], digest[2], digest[3], digest[4],
	digest[5], digest[6], digest[7], digest[8], digest[9],
	digest[10], digest[11], digest[12], digest[13], digest[14],
	digest[15]);
      nodesetprop (dev, "MD5_STATE", buf);
*/

      devtype = config[PCI_CLASS_DEVICE + 1] << 8 | config[PCI_CLASS_DEVICE];
      if (devtype == PCI_CLASS_SERIAL_USB)
      {
	/* Test to see if it's UHCI or OHCI */
	if (config[PCI_CLASS_PROG] == 0)
	{
	  nodesetprop (dev, "DRIVER", "usb-uhci");
	}
	else if (config[PCI_CLASS_PROG] == 0x10)
	{
	  nodesetprop (dev, "DRIVER", "usb-ohci");
	}
	else if (config[PCI_CLASS_PROG] == 0x20)
	{
	  nodesetprop (dev, "DRIVER", "ehci-hcd");
	}
      }
      if (devtype == PCI_CLASS_SERIAL_FIREWIRE)
      {
	/* Test to see if it's OHCI */
	if (config[PCI_CLASS_PROG] == 0x10)
	{
	  nodesetprop (dev, "DRIVER", "ohci1394");
	}
      }
      /* Check for an i2o device. Note that symbios controllers
       * also need i2o_scsi module. Dunno how to delineate that
       * here. */
      if (devtype == 0x0e00)
      {
	if ((config[PCI_CLASS_PROG] == 0 || config[PCI_CLASS_PROG] == 1)
	    && !strcmp (nodegetprop (dev, "DRIVER"), "unknown"))
	{
	  nodesetprop (dev, "DRIVER", "i2o_block");
	}
      }
      if (devtype == PCI_CLASS_BRIDGE_CARDBUS)
      {
	nodesetprop (dev, "DRIVER", "yenta_socket");
      }
      if (isDisabled (p, config))
      {
	nodesetprop (dev, "DRIVER", "disabled");
      }
      if (atoi (nodegetprop (dev, "PCIBUS")) == PCI_CARDBUS)
      {
	nodesetprop (dev, "DETACHED", "1");
      }
      /* Sure, reuse a pci id for an incompatible card. */
      if (atoi (nodegetprop (dev, "VENDORID")) == 0x10ec
	  && atoi (nodegetprop (dev, "DEVICEID")) == 8139
	  && config[PCI_REVISION_ID] >= 0x20)
      {
	nodesetprop (dev, "DRIVER", "8139cp");
      }

      if (strcmp (nodegetprop (dev, "DRIVER"), "unknown")
	  && strcmp (nodegetprop (dev, "DRIVER"), "disabled")
	  && strcmp (nodegetprop (dev, "DRIVER"), "ignore"))
      {
//      if (!type || (type < 0xff && (type == devtype >> 8))
//          || (type == kudzuToPci (pciToKudzu (devtype))))
	{
//          a_dev = nodealloc ();
//          nodesetprop (a_dev, "BUS", "PCI");
//        dev->class = pciToKudzu (devtype);
	  nodesetprop (dev, "CLASS", pciToKudzu (devtype));
	  switch (devtype)
	  {
	  case PCI_CLASS_NETWORK_TOKEN_RING:
	    nodesetprop (dev, "DEVICE", "tr");
	    break;
	  case PCI_CLASS_NETWORK_ETHERNET:
	  case PCI_CLASS_NETWORK_FDDI:
	  case PCI_CLASS_NETWORK_ATM:
	    nodesetprop (dev, "DEVICE", "eth");
	    break;
	  case PCI_CLASS_COMMUNICATION_OTHER:
	    checkPCISerial (dev, p);
	    break;
	  default:
	    break;
	  }
//        a_dev->index = order;
//        order++;

	  MD5Init (&context);
	  MD5UpdateS (&context, nodegetprop (dev, "VENDORID"));
	  MD5UpdateS (&context, nodegetprop (dev, "DEVICEID"));
	  MD5UpdateS (&context, nodegetprop (dev, "SUBVENDORID"));
	  MD5UpdateS (&context, nodegetprop (dev, "SUBDEVICEID"));
	  MD5Final (digest, &context);

	  sprintf (buf,
		   "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
		   digest[0], digest[1], digest[2], digest[3], digest[4],
		   digest[5], digest[6], digest[7], digest[8], digest[9],
		   digest[10], digest[11], digest[12], digest[13], digest[14],
		   digest[15]);

	  nodesetprop (dev, "MD5_ID", buf);

	  devices = nodeadd (devices, dev);
	}
      }
//      pciFreeDevice (dev);
    }
    pci_cleanup (pacc);
  }

  if (pciDeviceList && init_list)
    pciFreeDrivers ();

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
