#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "md5.h"
#include "list.h"
#include "module.h"

#if defined(__i386__) || defined(__alpha__)

#include "pcmcia/cs_types.h"
#include "pcmcia/cistpl.h"
#include "pcmcia/cs.h"
#include "pcmcia/bulkmem.h"
#include "pcmcia/ds.h"

char *
pcmciaToKudzu (unsigned int class)
{
  if (!class)
    return "UNSPEC";
  switch (class)
  {
  case CISTPL_FUNCID_NETWORK:
    return "NETWORK";
  case CISTPL_FUNCID_SCSI:
    return "SCSI";
  case CISTPL_FUNCID_SERIAL:
    return "MODEM";
  case CISTPL_FUNCID_FIXED:
    return "HD";
  default:
    return "OTHER";
  }
}

static int
get_tuple (int fd, unsigned char code, ds_ioctl_arg_t * arg)
{
  arg->tuple.DesiredTuple = code;
  arg->tuple.Attributes = TUPLE_RETURN_COMMON;
  arg->tuple.TupleOffset = 0;
  if ((ioctl (fd, DS_GET_FIRST_TUPLE, arg) == 0) &&
      (ioctl (fd, DS_GET_TUPLE_DATA, arg) == 0) &&
      (ioctl (fd, DS_PARSE_TUPLE, arg) == 0))
    return 0;
  else
    return -1;
}

static int pcmcia_major = 0;

static int
lookup_dev (char *name)
{
  FILE *f;
  int n;
  char s[32], t[32];

  f = fopen ("/proc/devices", "r");
  if (f == NULL)
    return -errno;
  while (fgets (s, 32, f) != NULL)
  {
    if (sscanf (s, "%d %s", &n, t) == 2)
      if (strcmp (name, t) == 0)
	break;
  }
  fclose (f);
  if (strcmp (name, t) == 0)
    return n;
  else
    return -ENODEV;
}

static int
open_sock (int sock)
{
  int fd;
  char fn[64];
  dev_t dev = (pcmcia_major << 8) + sock;

  snprintf (fn, 64, "/tmp/tmpdev-%d", getpid ());
  if (mknod (fn, (S_IFCHR | 0600), dev) == 0)
  {
    fd = open (fn, O_RDONLY);
    unlink (fn);
    if (fd >= 0)
      return fd;
  }
  return -1;
}

#define REMOVE	1
#define ADD	2
#define MEM	1
#define IO	2
#define IRQ	3

static struct adjust_t adj[] = {
  {ADD, IO, 0,.resource = {.io = {0x0c00, 0x0100, 0}}},
  {ADD, IO, 0,.resource = {.io = {0x0100, 0x0200, 0}}},
  {ADD, MEM, 0,.resource = {.memory = {0xc0000, 0x40000}}},
  {ADD, MEM, 0,.resource = {.memory = {0x60000000, 0x1000000}}},
  {ADD, MEM, 0,.resource = {.memory = {0xa0000000, 0x1000000}}},
  {ADD, IO, 0,.resource = {.io = {0x0a00, 0x0100, 0}}},
  {REMOVE, IRQ, 0,.resource = {.irq = {4}}},
  {REMOVE, IRQ, 0,.resource = {.irq = {7}}},
  {REMOVE, IRQ, 0,.resource = {.irq = {12}}},
  {0, 0, 0}
};

static void
adjust_resources (fd)
{
  int i = 0;

  for (; adj[i].Action; i++)
  {
    ioctl (fd, DS_ADJUST_RESOURCE_INFO, &adj[i]);
  }
}

static char *pcmciaDesc[] = {
  NULL,
  "Generic Memory Card",
  "Generic Serial/Modem Card",
  "Generic Parallel Port Card",
  "Generic ATA/IDE Disk",
  NULL,
  NULL,
  NULL,
  NULL,
};

static char *pcmciaDriver[] = {
  NULL,
  "memory_cs",
  "serial_cs",
  "parallel_cs",
  "ide_cs",
  NULL,
  NULL,
  NULL,
  NULL,
};
#endif

struct pcmciaDevice
{
  /* common fields */
  struct device *next;		/* next device in list */
  int index;
  char *device;			/* device file associated with it */
  char *driver;			/* driver to load, if any */
  char *desc;			/* a description */
  int detached;
  /* pcmcia-specific fields */
  unsigned int vendorId;	/* vendor id */
  unsigned int deviceId;	/* device id */
  unsigned int function;	/* logical function # */
  unsigned int slot;		/* slot the card is in */
};

struct pcmciaSearchDev
{
  struct pcmciaDevice pdev;
  char *vers_str[5];
};

static struct pcmciaSearchDev *pcmciaDeviceList = NULL;
static int numPcmciaDevices = 0;

static int
devCmpSort (const void *a, const void *b)
{
  const struct pcmciaSearchDev *one = a;
  const struct pcmciaSearchDev *two = b;
  int x = 0, y = 0, z = 0;
  int i;

  x = (one->pdev.vendorId - two->pdev.vendorId);
  y = (one->pdev.deviceId - two->pdev.deviceId);
  if (x)
    return x;
  else if (y)
    return y;
  for (i = 0; i < 5; i++)
  {
    if (one->vers_str[i] && two->vers_str[i])
      z = strcmp (one->vers_str[i], two->vers_str[i]);
    else
      z = one->vers_str[i] - two->vers_str[i];
    if (z)
      return z;
  }
  return 0;
}

static int
devCmpSearch (const void *a, const void *b)
{
  const struct pcmciaSearchDev *one = a;
  const struct pcmciaSearchDev *two = b;
  int x = 0, y = 0, z = 0;
  int i;

  x = (one->pdev.vendorId - two->pdev.vendorId);
  y = (one->pdev.deviceId - two->pdev.deviceId);
  if (x)
    return x;
  else if (y)
    return y;
  for (i = 0; i < 5; i++)
  {
    if (one->vers_str[i] && two->vers_str[i])
    {
      if (!strcmp (two->vers_str[i], "*"))
	z = 0;
      else
	z = strcmp (one->vers_str[i], two->vers_str[i]);
    }
    else
      z = one->vers_str[i] - two->vers_str[i];
    if (z)
      return z;
  }
  return 0;
}

int
pcmciaReadDrivers (char *filename)
{
  int fd;
  struct pcmciaSearchDev *nextDevice, *tmpdev = NULL, key;
  char *module, *m1, *m2;
  char *descrip = "";
  char *vers_str[5];
  char *buf, *start, *tmp, *next;
  int numDrivers, merge = 0, i;
  unsigned int vendid, devid;

  if (filename)
  {
    fd = open (filename, O_RDONLY);
    if (fd < 0)
      return -1;
  }
  else
  {
    fd = open ("/etc/pcmcia/config", O_RDONLY);
    if (fd < 0)
    {
      fd = open ("./config", O_RDONLY);
      if (fd < 0)
	return -1;
    }
  }
  buf = bufFromFd (fd);
  /* upper bound */
  numDrivers = 1;
  start = buf;
  do
  {
    if (start[0] == '\n')
      start++;
    if (!strncmp (start, "card \"", 5))
      numDrivers++;
  }
  while ((start = strchr (start, '\n')));

  if (pcmciaDeviceList)
    merge = 1;

  pcmciaDeviceList = realloc (pcmciaDeviceList, sizeof (*pcmciaDeviceList) *
			      (numPcmciaDevices + numDrivers));
  nextDevice = pcmciaDeviceList + numPcmciaDevices;

  memset (vers_str, '\0', sizeof (vers_str));
  vendid = devid = 0;
  m1 = m2 = NULL;

  start = buf;
  while (start && *start)
  {
    while (isspace (*start))
      start++;
    next = strchr (start, '\n');
    if (next)
      next++;
    if (!strncmp (start, "card \"", 6))
    {
      start += 6;
      tmp = strchr (start, '"');
      *tmp = '\0';
      tmp++;
      descrip = strdup (start);
      start = next;
      continue;
    }
    if (!strncmp (start, "version \"", 9))
    {
      start += 9;
      for (i = 0; i < 5; i++)
      {
	tmp = strchr (start, '"');
	*tmp = '\0';
	tmp++;
	vers_str[i] = strdup (start);
	start = tmp;
	if ((start = strchr (start, '"')) && start < (next - 1))
	{
	  start++;
	  continue;
	}
	else
	{
	  break;
	}
      }
      start = next;
      continue;
    }
    if (!strncmp (start, "manfid ", 7))
    {
      start += 7;
      vendid = strtoul (start, &start, 16);
      if (vendid && start)
      {
	start++;
	devid = strtoul (start, NULL, 16);
      }
      start = next;
      continue;
    }
    if (!strncmp (start, "bind \"", 6))
    {
      start += 6;
      tmp = strchr (start, '"');
      *tmp = '\0';
      tmp++;
      m1 = strdup (start);
      if ((start = strchr (tmp, ',')) && start < next - 1)
      {
	if ((start = strchr (start, '"')) && start < next - 1)
	{
	  start++;
	  tmp = strchr (start, '"');
	  *tmp = '\0';
	  m2 = strdup (start);
	}
      }
      if (m1 && m2)
      {
	module = malloc (strlen (m1) + strlen (m2) + 2);
	sprintf (module, "%s/%s", m1, m2);
	free (m1);
	free (m2);
      }
      else
      {
	module = m1;
      }

      if (merge)
      {
	tmpdev = nextDevice;
	key.pdev.vendorId = vendid;
	key.pdev.deviceId = devid;
	for (i = 0; i < 5; i++)
	{
	  key.vers_str[i] = vers_str[i];
	}
	nextDevice = bsearch (&key, pcmciaDeviceList,
			      numPcmciaDevices,
			      sizeof (struct pcmciaSearchDev), devCmpSort);

	if (!nextDevice)
	{
	  nextDevice = tmpdev;
	  tmpdev = NULL;
	  numPcmciaDevices++;
	}
	else
	{
	  if (nextDevice->pdev.device)
	    free (nextDevice->pdev.device);
	  if (nextDevice->pdev.desc)
	    free (nextDevice->pdev.desc);
	  if (nextDevice->pdev.driver)
	    free (nextDevice->pdev.driver);
	  for (i = 0; i < 5; i++)
	  {
	    if (nextDevice->vers_str[i])
	      free (nextDevice->vers_str[i]);
	  }
	}
      }
      else
	numPcmciaDevices++;
      nextDevice->pdev.vendorId = vendid;
      nextDevice->pdev.deviceId = devid;
      nextDevice->pdev.desc = strdup (descrip);
      nextDevice->pdev.next = NULL;
      nextDevice->pdev.device = NULL;
      nextDevice->pdev.driver = strdup (module);
      for (i = 0; i < 5; i++)
      {
	nextDevice->vers_str[i] = vers_str[i];
      }
      if (merge && tmpdev)
	nextDevice = tmpdev;
      else
      {
	nextDevice++;
	if (merge)
	  qsort (pcmciaDeviceList, numPcmciaDevices,
		 sizeof (*pcmciaDeviceList), devCmpSort);
      }

      free (module);
      free (descrip);
      m1 = m2 = NULL;
      memset (vers_str, '\0', sizeof (vers_str));
      vendid = devid = 0;
    }
    start = next;
  }
  qsort (pcmciaDeviceList, numPcmciaDevices, sizeof (*pcmciaDeviceList),
	 devCmpSort);
  return 0;
}

void
pcmciaFreeDrivers (void)
{
  int x, i;

  if (pcmciaDeviceList)
  {
    for (x = 0; x < numPcmciaDevices; x++)
    {
      if (pcmciaDeviceList[x].pdev.device)
	free (pcmciaDeviceList[x].pdev.device);
      if (pcmciaDeviceList[x].pdev.driver)
	free (pcmciaDeviceList[x].pdev.driver);
      if (pcmciaDeviceList[x].pdev.desc)
	free (pcmciaDeviceList[x].pdev.desc);
      for (i = 0; i < 5; i++)
	if (pcmciaDeviceList[x].vers_str[i])
	  free (pcmciaDeviceList[x].vers_str[i]);
    }
    free (pcmciaDeviceList);
    pcmciaDeviceList = NULL;
    numPcmciaDevices = 0;
  }
}
struct node *
probe (struct node *devices, int params)
{
  struct node *pcmciadev, *seconddev;
  int fd, ns, i, init_list = 0;
  ds_ioctl_arg_t arg;
  cistpl_vers_1_t *vers = &arg.tuple_parse.parse.version_1;
  cistpl_manfid_t *manfid = &arg.tuple_parse.parse.manfid;
  cistpl_funcid_t *funcid = &arg.tuple_parse.parse.funcid;
  cs_status_t status;
  char *temp;
  struct pcmciaSearchDev key, *searchdev;
  char buf[0x100];
  char tmpstr[256];
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "PCMCIA", 1);

#if defined(__i386__) || defined(__alpha__)

  if (!pcmciaDeviceList)
  {
    pcmciaReadDrivers (NULL);
    init_list = 1;
  }

  pcmcia_major = lookup_dev ("pcmcia");
  if (pcmcia_major < 0)
    goto out;
  for (ns = 0; ns < 8; ns++)
  {
    unsigned short vend = 0, dev = 0;
    unsigned char function=0;
    char *version_strs[5];
    int found = 0;
    char desc[256];

    fd = open_sock (ns);
    if (fd < 0)
      break;

    memset (version_strs, '\0', 5 * sizeof (char *));
    memset (desc, '\0', 256 * sizeof (char));

    adjust_resources (fd);
    /* Ignore cardbus cards */
    status.Function = 0;
    if (ioctl (fd, DS_GET_STATUS, &status) == 0)
    {
      if (status.CardState & CS_EVENT_CB_DETECT ||
	  !status.CardState & CS_EVENT_CARD_DETECT)
	continue;
    }
    else
      continue;

    if (get_tuple (fd, CISTPL_VERS_1, &arg) == 0)
    {
      for (i = 0; i < vers->ns; i++)
      {
	version_strs[i] = strdup (vers->str + vers->ofs[i]);
      }
      version_strs[i] = NULL;
      found++;
    }
    if (get_tuple (fd, CISTPL_MANFID, &arg) == 0)
    {
      vend = manfid->manf;
      dev = manfid->card;
      found++;
    }
    if (get_tuple (fd, CISTPL_FUNCID, &arg) == 0)
    {
      function = funcid->func;
      found++;
    }

    pcmciadev = nodealloc ();
    nodesetprop (pcmciadev, "BUS", "PCMCIA");
    sprintf (buf, "pcmcia.%d.0", ns);
    nodesetprop (pcmciadev, "CONN", buf);
    nodesetpropint (pcmciadev, "SLOT", ns);
    nodesetpropint (pcmciadev, "VENDORID", vend);
    nodesetpropint (pcmciadev, "DEVICEID", dev);

    MD5Init (&context);
    sprintf (buf, "pcmcia %d %d 0", vend, dev);
    MD5UpdateS (&context, buf);
    MD5Final (digest, &context);

    sprintf (buf,
	     "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
	     digest[0], digest[1], digest[2], digest[3], digest[4],
	     digest[5], digest[6], digest[7], digest[8], digest[9],
	     digest[10], digest[11], digest[12], digest[13],
	     digest[14], digest[15]);

    nodesetprop (pcmciadev, "MD5_ID", buf);

    memset (&key, '\0', sizeof (key));
    key.pdev.vendorId = vend;
    key.pdev.deviceId = dev;
    for (i = 0; i < 5; i++)
      key.vers_str[i] = version_strs[i];
    for (i = 4; i > 0; i--)
    {
      key.vers_str[i] = NULL;
      searchdev = bsearch (&key, pcmciaDeviceList,
			   numPcmciaDevices,
			   sizeof (struct pcmciaSearchDev), devCmpSearch);
      if (searchdev)
	break;
    }
    if (!searchdev)
    {
      memset (&key, '\0', sizeof (key));
      for (i = 0; i < 5; i++)
	key.vers_str[i] = version_strs[i];
      for (i = 4; i > 0; i--)
      {
	key.vers_str[i] = NULL;
	searchdev = bsearch (&key, pcmciaDeviceList,
			     numPcmciaDevices,
			     sizeof (struct pcmciaSearchDev), devCmpSearch);
	if (searchdev)
	  break;
      }
    }
    if (!searchdev)
    {
      memset (&key, '\0', sizeof (key));
      key.pdev.vendorId = vend;
      key.pdev.deviceId = dev;
      searchdev = bsearch (&key, pcmciaDeviceList,
			   numPcmciaDevices,
			   sizeof (struct pcmciaSearchDev), devCmpSearch);
    }

    if (searchdev)
    {
//      po co to szukanie????
//      pcmciadev = pcmciaNewDevice (&searchdev->pdev);
    }
    else
    {
      switch (function)
      {
      case CISTPL_FUNCID_MEMORY:
      case CISTPL_FUNCID_SERIAL:
      case CISTPL_FUNCID_PARALLEL:
      case CISTPL_FUNCID_FIXED:

	nodesetprop (pcmciadev, "CLASS", pcmciaToKudzu (function));
	nodesetprop (pcmciadev, "DESC", pcmciaDesc[function]);
	nodesetprop (pcmciadev, "DRIVER", pcmciaDriver[function]);
	break;
      default:
	memset (tmpstr, '\0', sizeof (tmpstr));
	if (version_strs[0])
	{
	  strcat (tmpstr, version_strs[0]);
	  for (i = 1; i < 5; i++)
	  {
	    if (version_strs[i])
	    {
	      strcat (tmpstr, " ");
	      strcat (tmpstr, version_strs[i]);
	    }
	  }
	}
	else
	{
	  nodesetprop (pcmciadev, "DESC", "Unknown PCMCIA Card");
	}
	nodesetprop (pcmciadev, "DESC", tmpstr);
	nodesetprop (pcmciadev, "DRIVER", "unknown");
	break;
      }
    }

    nodesetprop (pcmciadev, "FUNCTION", "0");
    nodesetprop (pcmciadev, "CLASS", pcmciaToKudzu (function));
    nodesetpropint (pcmciadev, "SLOT", ns);
    nodesetpropint (pcmciadev, "VENDORID", vend);
    nodesetpropint (pcmciadev, "DEVICEID", dev);

    if ((temp = strchr (nodegetprop (pcmciadev, "DRIVER"), '/')))
    {
      *temp = '\0';
      seconddev = nodeclone (pcmciadev);

      sprintf (buf, "pcmcia.%d.1", ns);
      nodesetprop (pcmciadev, "CONN", buf);

      MD5Init (&context);
      sprintf (buf, "pcmcia %d %d 1", vend, dev);
      MD5UpdateS (&context, buf);
      MD5Final (digest, &context);

      nodesetprop (seconddev, "FUNCTION", "1");
      nodesetprop (seconddev, "DRIVER", temp + 1);
      /* fixup */
      if (!strcmp (nodegetprop (seconddev, "DRIVER"), "xirc2ps_cs") ||
	  !strcmp (nodegetprop (seconddev, "DRIVER"), "3c574_cs") ||
	  !strcmp (nodegetprop (seconddev, "DRIVER"), "3c589_cs") ||
	  !strcmp (nodegetprop (seconddev, "DRIVER"), "pcnet_cs") ||
	  !strcmp (nodegetprop (seconddev, "DRIVER"), "smc91c92_cs"))
	nodesetprop (seconddev, "CLASS", "NETWORK");
      if (!strcmp (nodegetprop (pcmciadev, "DRIVER"), "xirc2ps_cs") ||
	  !strcmp (nodegetprop (pcmciadev, "DRIVER"), "3c574_cs") ||
	  !strcmp (nodegetprop (pcmciadev, "DRIVER"), "3c589_cs") ||
	  !strcmp (nodegetprop (pcmciadev, "DRIVER"), "pcnet_cs") ||
	  !strcmp (nodegetprop (pcmciadev, "DRIVER"), "smc91c92_cs"))
	nodesetprop (pcmciadev, "CLASS", "NETWORK");

      devices = nodeadd (devices, seconddev);
    }
    devices = nodeadd (devices, pcmciadev);

  }
out:
  if (pcmciaDeviceList && init_list)
    pcmciaFreeDrivers ();
#endif

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
